#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from __future__ import with_statement

from ionas.fileaccess import (MODE_READ, MODE_WRITE, MODE_APPEND,
                                DEFAULT_LOCALE,)
from ionas.exception import *
from ionas.fileaccess.filelikeobj import (FileLikeObj,)


import os
import sys
import stat
import errno
import locale
import httplib2
import time

from ionas.exception import *
from landiskaccess.dropboxaccess import (DropboxAccess,)
from landiskaccess.dropboxclient import (InsufficientSpaceDropboxError,
                                            NotFoundDropboxError,
                                            ConflictDropboxError,
                                            ServerDropboxError,
                                            IncompleteReadDropboxError,
                                            BaseDropboxError)
from pkgaddon.dropbox_config import (APP_KEY, APP_SECRET, SESSION_DIR,)


BUF_SIZE = 65536

RETRY_COUNT = 3


def dropbox_except_handler(path):
    exc_ins = sys.exc_info()[1]
    if isinstance(exc_ins, InsufficientSpaceDropboxError):
        raise NoSpaceError(exc_ins, path)
    if isinstance(exc_ins, NotFoundDropboxError):
        raise NoEntryError(exc_ins, path)
    if isinstance(exc_ins, ConflictDropboxError):
        raise ExistsError(exc_ins, path)
    raise


def to_open_flag(mode):
    if mode == MODE_READ:
        return "r"
    if mode == MODE_WRITE:
        return "w"
    if mode == MODE_APPEND:
        return "w"
    return "r"


class DropboxFileObj(FileLikeObj):
    
    def __renew_file(self):

        count = 0
        while True:
            try:
                if self.file != None:
                    self.file.close()

                self.file = self.accessor.get_renewed_fileobj(
                                self.path, self.mode, self.length)
                return
            except:
                try:
                    dropbox_except_handler(self.path)
                except (RuntimeError, ValueError):
                    if count > RETRY_COUNT:
                        raise
                except (PermissionError, NoSpaceError, NoEntryError):
                    raise
                except:
                    pass

                if count > RETRY_COUNT:
                    raise
                count += 1

    def __call_and_retry_with_renew(self, func, file_like=None):
        count = 0
        while True:
            try:
                return func()
            except:
                try:
                    dropbox_except_handler(self.path)
                except (RuntimeError, ValueError):
                    if count > RETRY_COUNT:
                        raise

                except (PermissionError, NoSpaceError, NoEntryError):
                    raise
                except:
                    pass

                if count > RETRY_COUNT:
                    raise

                self.__renew_file()
                if hasattr(file_like, "seek"):
                    file_like.seek(0)

                count += 1
                time.sleep(10)

    def __init__(self, file, path, mode, accessor, length):

        FileLikeObj.__init__(self, file, path)
        self.mode = mode
        self.accessor = accessor
        self.length = length

    def __len__(self):

        return int(self.accessor.stat(self.path)[stat.ST_SIZE])

    def __read(self, size):

        self.point += size
        return self.file.read(size)

    def __write(self, file_like):

        while True:
            try:
                buf = file_like.read(BUF_SIZE)
            except:
                dropbox_except_handler(self.path)

            if not buf:
                break
            self.file.write(buf)

    def __read_by_path(self, file_like):

        try:
            file_like.write(self.file)
        except:
            dropbox_except_handler(self.path)

    def __write_by_path(self, file_like):

        try:
            # FIXME 一度に書いて大丈夫か?
            self.__write(file_like)
        except:
            dropbox_except_handler(self.path)

    def __close(self):

        try:
            self.file.close()
        except:
            dropbox_except_handler(self.path)

    def read(self, size):

        func = lambda: self.__read(size)
        return self.__call_and_retry_with_renew(func)

    def write(self, file_like):

        if self.file == None:
            open_mode = to_open_flag(self.mode)
            try:
                self.length=file_like.get_stat()[stat.ST_SIZE]
            except:
                self.length=len(file_like)

            self.__renew_file()

        func = lambda: self.__write(file_like)
        return self.__call_and_retry_with_renew(func, file_like)

    def write_by_path(self, file_like):

        if self.file == None:
            open_mode = to_open_flag(self.mode)
            try:
                self.length=file_like.get_stat()[stat.ST_SIZE]
            except:
                self.length=len(file_like)

            self.__renew_file()

        func = lambda: self.__write_by_path(file_like)
        return self.__call_and_retry_with_renew(func, file_like)

    def read_by_path(self, file_like):

        func = lambda: self.__read_by_path(file_like)
        return self.__call_and_retry_with_renew(func, file_like)

    def get_stat(self):

        return self.access.stat(self.path)

    def close(self):

        func = lambda: self.__close()
        return self.__call_and_retry_with_renew(func)


class BackupDropboxAccess():

    def __call_and_retry(self, func, path):
        count = 0
        while True:
            try:
                return func()
            except:
                try:
                    dropbox_except_handler(path)
                except (RuntimeError, ValueError):
                    if count > RETRY_COUNT:
                        raise

                except (PermissionError, NoSpaceError, NoEntryError):
                    raise
                except:
                    pass

                if count > RETRY_COUNT:
                    raise

                # Retry
                count += 1
                self._reset_access()
                time.sleep(10)

    def _reset_access(self):
        """Reset access."""

        try:
            del self.access
        except:
            pass

        connection_info = self.target_info.connection_info
        self.access = DropboxAccess(
                        APP_KEY,
                        APP_SECRET,
                        session_path=self.session_path,
                        proxy=self.proxy)

        if not os.path.exists(self.session_path):
            self.access.authenticate({"code": connection_info.code})
        else:
            self.access.authenticate()

    def get_renewed_fileobj(self, path, mode, length):
        """Renew file obj and return it"""

        open_mode = to_open_flag(mode)
        try:
            return self.access.open(path, open_mode, length=length)
        except:
            self._reset_access()
            time.sleep(10)
            return self.access.open(path, open_mode, length=length)

    def __init__(self, target_info, proxy=None, encoding=""):
        self.sep = "/"
        
        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE

        self.target_info = target_info
        self.proxy = proxy

        connection_info = target_info.connection_info
        self.session_path = os.path.join(SESSION_DIR, connection_info.name) + \
                                            ".session"
        self.access = DropboxAccess(
                        APP_KEY,
                        APP_SECRET,
                        session_path=self.session_path,
                        proxy=self.proxy)
        if not os.path.exists(self.session_path):
            self.access.authenticate({"code": connection_info.code})
        else:
            self.access.authenticate()


    def listdir(self, path):
        if isinstance(path, str):
            path = path.decode(self.encoding)
        func = lambda: self.access.get_contents(path)
        contents_list = []
        for key in self.__call_and_retry(func, path).keys():
            if isinstance(key, unicode):
                contents_list.append(key.encode(self.encoding))
            else:
                contents_list.append(key)
        return contents_list

    def isdir(self, entry):  # FIXME 高速化するなら必要かも
        pass

    def open(self, path, mode):
        if isinstance(path, str):
            path = path.decode(self.encoding)
        open_mode = to_open_flag(mode)
        opened_file = None
        if open_mode == "r":
            try:
                opened_file = self.access.open(path, open_mode)
            except:
                self._reset_access()
                time.sleep(10)
                opened_file = self.access.open(path, open_mode)

        func = lambda: DropboxFileObj(opened_file,
                    path, mode, accessor=self, length=-1)
        return self.__call_and_retry(func, path)

    def stat(self, path):

        if isinstance(path, str):
            path = path.decode(self.encoding)
        func = lambda: self.access.stat_file(path)
        stat_info = self.__call_and_retry(func, path)
        try:
            if stat.ST_MTIME in stat_info:
                stat_info[stat.ST_MTIME] = stat_info[stat.ST_MTIME] + \
                                            9 * 3600
        except:
            pass
        return stat_info

    def chmod(self, path, mode):
        return NotImplementedError

    def utime(self, path, mtime):
        return NotImplementedError

    def remove(self, path):
        if isinstance(path, str):
            path = path.decode(self.encoding)
        func = lambda: self.access.delete_file(path)
        self.__call_and_retry(func, path)

    def rmdir(self, path):
        if isinstance(path, str):
            path = path.decode(self.encoding)
        func = lambda: self.access.delete_directory(path)
        self.__call_and_retry(func, path)

    def mkdir(self, path):
        if isinstance(path, str):
            path = path.decode(self.encoding)
        func = lambda: self.access.make_directory(path)
        self.__call_and_retry(func, path)

    def move(self, src, dst):
        if isinstance(src, str):
            src = src.decode(self.encoding)
        if isinstance(dst, str):
            dst = dst.decode(self.encoding)

        func = lambda: self.access.move_directory(src, dst)
        self.__call_and_retry(func, dst)
